@@ -11,7 +11,7 @@ import argparse
 import sys
 import os
 import shellescape
-from PIL import Image
+from PIL import Image, ImageSequence
 import PIL
 from io import BytesIO
 from urllib.parse import quote
@@ -19,6 +19,7 @@ import traceback
 import subprocess
 import json
 import webcolors
+import emoji
 from ast import literal_eval as make_tuple
 import bs4
 from bs4 import BeautifulSoup
@@ -27,20 +28,32 @@ import math as math_module
 import shutil
 import typing
 import html
+import uuid
 import warnings
 from .latex2svg import latex2svg
 from .helpers import heading_name_to_id_value
+try:
+    import tidylib
+    imported_tidylib = True
+except (ImportError, ModuleNotFoundError):
+    imported_tidylib = False
+try:
+    import difflib
+    imported_difflib = True
+except (ImportError, ModuleNotFoundError):
+    imported_difflib = False
 
 MODULE_PATH = os.path.join(*os.path.split(__file__)[:-1])
-DEBUG = False  # weather to print debug information
+DEBUG = False  # whether to print debug information
+DEBUG_HASHES = False
 HASH_FUNCTION_TO_USE_ON_IMAGES = lambda x: hashlib.md5(x.encode() if type(x) is str else x)
 
 
-def open_local(path, mode):
-    return open(os.path.join(MODULE_PATH, path), mode)
+def open_local(path, *args, **kwargs):
+    return open(os.path.join(MODULE_PATH, path), *args, **kwargs)
 
 
-HELP = open_local("help.txt", "r").read()
+HELP = open_local("help.txt", "r", encoding="utf-8").read()
 
 
 def is_not_escaped(s: str, i: int):
@@ -74,7 +87,94 @@ def get_fitting_replacement(thing_to_replace: str, table_of_replacement: dict, t
     return replacement
 
 
-def find_and_replace_formulas_in_markdown(md: str, support_formulas=True):
+def shortcode_to_emoji(shortcode: str, emoji_support_level):
+    if emoji_support_level == 0:
+        return shortcode
+    else:
+        if "." not in shortcode:
+            # we take this as a hint that this is a non-custom emoji:
+            if shortcode not in emoji.EMOJI_ALIAS_UNICODE_ENGLISH:
+                warnings.warn("`" + shortcode + "` is not a valid emoji shortcode.\n"
+                              + "This does no harm; we are just letting you know in case you intended it to be an emoji"
+                              + " shortcode and mistyped it accidentally."
+                              + "")
+            return emoji.emojize(shortcode, use_aliases=True, variant="emoji_type")
+        else:
+            # we take this as a hint that this is a custom emoji:
+            escaped_shortcode = shortcode[1:-1].replace("'", "\'").replace('"', '\"')
+            emoji_name = escaped_shortcode.split("/")[-1].split("?")[0].split("&")[0]
+            title = emoji_name.rsplit(".", 1)[0].replace("-", " ").replace("_", " ")
+            for i in range(len(title)-1, -1, -1):
+                if title[i].isupper():
+                    title = title[:i] + " " + title[i:]
+            return ("<img src='" + escaped_shortcode + "' "
+                    + "title=':" + emoji_name + ":' "
+                    + 'alt="' + title + '" style="width:1em; height:1em;" is_emoji="true">')
+
+
+STD_EMOJI_CHARS = "-" + "_" + string.ascii_uppercase + string.ascii_lowercase + string.digits
+EXT_EMOJI_CHARS = "-._~/?#[]@!$&'()*+,;=" + string.ascii_uppercase + string.ascii_lowercase + string.digits  # <- no :
+EMOJI_WHITESPACE = ("\n", "\t", " ")
+
+
+def proceed_emoji_parsing(line, i, emoji_replacements, emoji_start, emoji_state, md, support_custom_emojis=False):
+    # values for emoji_state:
+    #  0: not within an emoji and not a whitespace
+    #  1: in whitespace
+    #  2: at : in front of an emoji
+    #  3: within text of emoji
+    #  4: at closing : of emoji
+    success = False
+    if emoji_state == 0:  # <- was not within an emoji and not a whitespace
+        if line[i] in EMOJI_WHITESPACE:
+            emoji_state = 1
+    elif emoji_state == 1:  # <- was in whitespace
+        if line[i] == ":" and is_not_escaped(line, i):
+            emoji_state = 2
+            emoji_start = i
+        elif line[i] not in EMOJI_WHITESPACE:
+            emoji_state = 0
+    elif emoji_state == 2:  # <- was at a : sign
+        if line[i] in (STD_EMOJI_CHARS if not support_custom_emojis else EXT_EMOJI_CHARS):
+            emoji_state = 3
+        elif line[i] in EMOJI_WHITESPACE:
+            emoji_state = 1
+        else:
+            emoji_state = 0
+    elif emoji_state == 3:  # <- was within an emoji
+        if line[i] in (STD_EMOJI_CHARS if not support_custom_emojis else EXT_EMOJI_CHARS):
+            emoji_state = 3
+        elif line[i] == ":" and support_custom_emojis and (i != len(line)-1) and (line[i+1] in EXT_EMOJI_CHARS + ":"):
+            emoji_state = 3  # <- support : in extended emojis
+        elif line[i] in EMOJI_WHITESPACE:
+            emoji_state = 1
+        elif line[i] == ":":
+            if i == len(line) - 1:
+                i += 1
+                success = True
+            else:
+                emoji_state = 4
+        else:
+            emoji_state = 0
+    elif emoji_state == 4:  # <- was at the trailing : of an emoji
+        if line[i] in EMOJI_WHITESPACE:
+            success = True
+        else:
+            emoji_state = 0
+    if success:
+        emoji_state = 1
+        emoji_shortcode = line[emoji_start:i]
+        replacement = get_fitting_replacement(emoji_shortcode, emoji_replacements, md, "e")
+        emoji_replacements[replacement] = emoji_shortcode
+        line_new = line[:emoji_start] + replacement + line[i:]
+        i += len(line_new) - len(line)
+        line = line_new
+        if i == len(line):
+            i -= 1
+    return line, i, emoji_replacements, emoji_start, emoji_state
+
+
+def find_and_replace_formulas_in_markdown(md: str, support_formulas=True, support_custom_emojis=False):
     """Takes markdown as a string and returns the markdown, but every formula is replaced with a random string, as well
     as a dict to translate these strings back to the formulas. This is done to evade the problem that special characters
     in formulas should not be escaped and that they should not be interpreted, e.g. as code etc.
@@ -90,8 +190,13 @@ def find_and_replace_formulas_in_markdown(md: str, support_formulas=True):
     special_characters_in_code = dict()
     headings = list()
     inside_multiline_code = False
+    # specifically for emojis:
+    emoji_replacements = dict()
+
     # iterate over the document's lines:
     for l_num in range(len(md_lines)):
+        emoji_state = 1  # 1 because the -1st letter of ever line is considered to be whitespace.
+        emoji_start = 0
         line = md_lines[l_num]
         # switch between multiline code blocks and things that aren't multiline code blocks:
         if line.strip().startswith("```"):
@@ -99,17 +204,36 @@ def find_and_replace_formulas_in_markdown(md: str, support_formulas=True):
             continue
         # we need to look for formulas, inline code blocks and table of content indicators outside:
         if not inside_multiline_code:
+            if DEBUG:
+                print(line)
             inside_inline_code = False
             formula_start = 0
             in_formula = False
-            # are we add a heading? if yes, register that and move on to the next line.
+            # are we at a heading? if yes, register that and move on to the next line.
             if line.strip().startswith("#") and " " in line and line.split(" ")[0] == "#" * len(line.split(" ")[0]):
+                # search for emojis:
+                i = -1
+                if DEBUG:
+                    print("\n" + line)
+                while i < len(line) - 1:
+                    i += 1
+                    if line[i] == "`" and is_not_escaped(line, i):
+                        inside_inline_code = not inside_inline_code
+                    if not inside_inline_code:
+                        line, i, emoji_replacements, emoji_start, emoji_state = proceed_emoji_parsing(
+                            line, i, emoji_replacements, emoji_start, emoji_state, md, support_custom_emojis)
+
                 headings.append((len(line.split(" ")[0]), line.split(" ", 1)[1]))
+                md_lines[l_num] = line
                 continue
             # we simply iterate over all characters of the line now, setting and unsetting flags as we pass them.
             i = -1
             while i < len(line) - 1:
                 i += 1
+                # do emoji checks:
+                if not inside_inline_code and not in_formula:
+                    line, i, emoji_replacements, emoji_start, emoji_state = proceed_emoji_parsing(
+                        line, i, emoji_replacements, emoji_start, emoji_state, md, support_custom_emojis)
                 # check whether a formula starts or ends here (only if formulas support is activated):
                 if support_formulas and line[i] == "$" and is_not_escaped(line, i) and not inside_inline_code:
                     if not in_formula:
@@ -139,8 +263,12 @@ def find_and_replace_formulas_in_markdown(md: str, support_formulas=True):
                 # handle escaped formula-signs if formula support is activated:
                 if support_formulas:
                     line = line.replace("\\$", "$")
+                if DEBUG:
+                    print(emoji_state, end="")
             # encode umlauts outside of inline code blocks:
             line = str(line.encode('ascii', 'xmlcharrefreplace'), encoding="utf-8")
+            if DEBUG:
+                print("\n")
         # search for non-ascii characters if we are in a multiline code block:
         else:
             for character in set(line):
@@ -151,7 +279,7 @@ def find_and_replace_formulas_in_markdown(md: str, support_formulas=True):
                     line = line.replace(character, replacement)
         md_lines[l_num] = line
 
-    return "\n".join(md_lines), formulas, special_characters_in_code, headings
+    return "\n".join(md_lines), formulas, special_characters_in_code, headings, emoji_replacements
 
 
 def header_name_to_link_to_header(header_name: str) -> str:
@@ -221,8 +349,8 @@ def render_toc(md_content: str, headings: typing.List[typing.Tuple[int, str]]) -
 
 # a stand-alone function to render tocs in markdown files in case one wants to do just that:
 
-def render_toc_in_md_file_stand_alone(md_content: str) -> str:
-    _, _, _, headers = find_and_replace_formulas_in_markdown(md_content)
+def render_toc_in_md_file_stand_alone(md_content: str, support_custom_emojis=False) -> str:
+    _, _, _, headers, _ = find_and_replace_formulas_in_markdown(md_content, support_custom_emojis)
     return render_toc(md_content, headers)
 
 
@@ -289,7 +417,8 @@ def formula2svg(formula, amount_of_svg_formulas):
 # Find and render formulas in the html code, and replace them correctly:
 
 
-def find_and_render_formulas_in_html(html_text: str, formulas: dict, special_characters_in_code: dict):
+def find_and_render_formulas_in_html(html_text: str, formulas: dict, special_characters_in_code: dict,
+                                     emoji_replacements: dict, emoji_support: int):
     """Takes some html (generated from markdown by the online github API) and a dictionary which maps a number of
     sequences to a number of formulas, and replaces each sequence with a LaTeX-rendering of the corresponding formula.
     The third parameter is a dictionary mapping replacements to special characters for use in code blocks.
@@ -312,6 +441,15 @@ def find_and_render_formulas_in_html(html_text: str, formulas: dict, special_cha
             special_character
         )
 
+    # replace emojis:
+    for sequence, emoji_shortcode in emoji_replacements.items():
+        if DEBUG:
+            print("replace:", sequence, shortcode_to_emoji(emoji_shortcode, emoji_support))
+        html_text = html_text.replace(
+            sequence,
+            shortcode_to_emoji(emoji_shortcode, emoji_support)
+        )
+
     return html_text
 
 
@@ -469,16 +607,25 @@ def find_fitting_hash_function(amount_of_images):
 # Hash an image:
 
 
-def hash_image(img):
+def hash_image(img, return_unhashed=False):
     if type(img) in (str, bytes):
         return HASH_FUNCTION_TO_USE_ON_IMAGES(img)
 
     pixel_data = list()
-    for pixel in list(img.getdata()):
-        pixel_data += list(pixel)
+    frames_durations = list()
+    for frame in ImageSequence.Iterator(img):
+        frame = frame.convert("RGBA")
+        if "duration" in frame.info:
+            frames_durations.append(str(frame.info["duration"]))
+        for pixel in list(frame.getdata()):
+            pixel_data += list(pixel)
     pixel_data_string = str(bytes(pixel_data), encoding="iso-8859-1")
-    pixel_data_string += "||" + str(img.size) + "||" + (str(img.format.lower() if img.format else None))
+    pixel_data_string += "||" + str(img.size) + "||" + (str(img.format.lower() if img.format else None)) + (
+        ("||" + str(",".join(frames_durations))) if frames_durations else ""
+    )
 
+    if return_unhashed:
+        return pixel_data_string
     return HASH_FUNCTION_TO_USE_ON_IMAGES(pixel_data_string)
 
 
@@ -501,13 +648,26 @@ def make_unused_name(base_file_name, file_name_addition, already_used_filenames,
     If ending is specified (in the form .y), the generated file name will be foo.bar.y instead of foo.bar.x."""
     def make_final_filename(name, add):
         return name.rsplit(".", 1)[0] + add + "." + name.rsplit(".", 1)[1]
+    def resulting_name_too_long(name, add):
+        return len(name) + len(add) > 260
+    def resulting_name_already_used(name, add):
+        return make_final_filename(name, add) in already_used_filenames
+
     if hash_of_image in hashes_to_filenames:
+        if DEBUG_HASHES:
+            print("image already present as", hashes_to_filenames[hash_of_image])
         return hashes_to_filenames[hash_of_image]
-    if not ending:
+    if ending is None:
         ending = "." + base_file_name.rsplit(".", 1)[-1]
     base_file_name = base_file_name.rsplit(".", 1)[0] + ending
-    while make_final_filename(base_file_name, file_name_addition) in already_used_filenames:
-        base_file_name = make_final_filename(base_file_name, "_")
+    while (
+            resulting_name_already_used(base_file_name, file_name_addition)
+            or resulting_name_too_long(base_file_name, file_name_addition)
+    ):
+        if resulting_name_already_used(base_file_name, file_name_addition):
+            base_file_name = make_final_filename(base_file_name, "_")
+        if resulting_name_too_long(base_file_name, file_name_addition):
+            base_file_name = base_file_name[:32] + "-" + uuid.uuid4().hex + "." + ending
     base_file_name = make_final_filename(base_file_name, file_name_addition)
     already_used_filenames.add(base_file_name)
     hashes_to_filenames[hash_of_image] = base_file_name
@@ -570,7 +730,10 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
          formulas_supporting_darkreader=False, extra_css=None,
          core_converter: typing.Union[str, typing.Callable] = markdown_to_html_via_github_api,
          compress_images=False, enable_image_downloading=True, box_width=None, toc=False, dont_make_images_links=False,
-         soft_wrap_in_code_boxes=False, suppress_online_fallbacks=False):
+         soft_wrap_in_code_boxes=False, suppress_online_fallbacks=False, validate_html=False, emoji_support=1):
+    # check emoji_support parameter:
+    if emoji_support not in (0, 1, 2):
+        raise Exception("--emoji-support must be one of 0, 1 and 2.")
     # set all to defaults:
     style_pdf = str2bool(style_pdf)
     math = str2bool(math)
@@ -614,8 +777,11 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
         print("\n------------\nOriginal content:\n------------\n\n", md_content)
 
     # replace formulas with random sequences and get a dict to map them back:
-    md_content, formula_mapper, special_chars_in_code_blocks, headings = find_and_replace_formulas_in_markdown(
-        md_content, math)
+    support_custom_emojis = (emoji_support >= 2)
+    md_content, formula_mapper, special_chars_in_code_blocks, headings, emoji_replacements = (
+        find_and_replace_formulas_in_markdown(md_content, math, support_custom_emojis))
+    if DEBUG:
+        print("emoji replacements:", emoji_replacements)
 
     # fail if we need to convert formulas, don't have the necessary dependencies and are suppressing online fallbacks:
     if suppress_online_fallbacks and formula_mapper and raw_formula2svg_online == raw_formula2svg:
@@ -672,7 +838,8 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
         html_content = html_bs4.__str__()
 
     # re-insert formulas in html, this time as proper svg images:
-    html_content = find_and_render_formulas_in_html(html_content, formula_mapper, special_chars_in_code_blocks)
+    html_content = find_and_render_formulas_in_html(html_content, formula_mapper, special_chars_in_code_blocks,
+                                                    emoji_replacements, emoji_support)
 
     if DEBUG:
         print("\n------------\nHtml content (with properly rendered formulas):\n------------\n\n", html_content)
@@ -732,9 +899,19 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
             image_src = original_markdown_image_src = img_soup_representation.get("src")
             if image_src == "":
                 continue  # <-- In case some images with no source where injected for some reason
-            save_image_as = re.split("[/\\\]", image_src)[-1]  # <--  take only the last element of the path
+            if img_soup_representation.has_attr("data-canonical-src"):
+                # work around for GitHub's image caching, which results in absurdly long image names.
+                save_image_as = img_soup_representation.get("data-canonical-src")
+            else:
+                save_image_as = image_src
+            if DEBUG:
+                print("new image_src:", save_image_as)
+            save_image_as = save_image_as.split("?")[0]  # <-- remove the extra url parts
+            save_image_as = re.split("[/\\\]", save_image_as)[-1]  # <--  take only the last element of the path
             save_image_as = save_image_as.rsplit(".", 1)[0]  # <-- remove the extension
             save_image_as = re.sub(r'(?u)[^-\w.]', '', save_image_as)  # <-- remove disallowed characters
+            if DEBUG:
+                print("-> original_save_image_s:", save_image_as)
             if image_src.startswith("./"):
                 image_src = image_src[2:]
 
@@ -809,17 +986,40 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
             # ensure we use no image name twice & finally save the image:
             save_image_as = make_unused_name(save_image_as + extension, "", saved_image_names, hashes_to_images,
                                              hash_image(img_object))  # <-- file name to save as
+            if DEBUG:
+                print("-> save_image_as:", save_image_as)
+                print("")
             cached_image_path = os.path.join(abs_image_paths, save_image_as)  # <-- path where we save it
             location_of_full_sized_image = image_name_to_image_src(save_image_as)  # <-how we call that path in the html
             if extension != ".svg":
-                img_object.save(cached_image_path)
+                # if extension == ".gif":
+                #     print(cached_image_path)
+                img_object.save(cached_image_path, save_all=(extension == ".gif"))
             else:
                 with open(cached_image_path, "wb") as img_out_file:
                     img_out_file.write(img_object)
 
+            # Check if hashing worked correctly:
+            if DEBUG_HASHES:
+                import time
+                t = time.time()
+                if extension != ".svg":
+                    hash1 = hash_image(Image.open(cached_image_path), return_unhashed=True)
+                else:
+                    hash1 = hash_image(open(cached_image_path, "rb").read(), return_unhashed=True)
+                hash2 = hash_image(img_object, return_unhashed=True)
+                if hash1 != hash2:
+                    warnings.warn(
+                        "image " + cached_image_path + " hashed incorrectly (not dramatic, but you cans till raise an\
+                        issue for this)."
+                        + (("hash difference:\n" + "".join(difflib.ndiff(hash1, hash2))) if imported_difflib else "")
+                        + "\n"
+                    )
+                print("time to compare hashes:", time.time() - t)
+
             # Open the final image and do compression, if it was specified to do so:
             height = None
-            if compression_information and extension != ".svg":
+            if compression_information and extension not in (".svg", ".gif"):
                 full_image = Image.open(cached_image_path)
                 # Determine the images' width if any is specified:
                 width = (
@@ -832,6 +1032,9 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
                     if img_soup_representation.has_attr("height") and img_soup_representation["height"].endswith("px")
                     else None
                 )
+                if img_soup_representation.has_attr("is_emoji") and img_soup_representation["is_emoji"] == "true":
+                    height = 128
+                    width = 128
                 if height and not width:
                     width = math_module.ceil(height * full_image.width / full_image.height)
                 # If no size is specified and srcset is set, generate a set of resolutions:
@@ -910,16 +1113,32 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
     if DEBUG:
         print("\n------------\nHtml with image links:\n------------\n\n", html_rendered)
 
-    # ensure <a href="#fufu"> gets converted to <a href="#user-content-fufu">:
+    # Add "user-content-" to anchors and internal links.
+
     contains_file_internal_links = False
     html_soup = BeautifulSoup(html_rendered, 'html.parser')
+    for element_soup_representation in html_soup.select("[id]"):
+        id_name = element_soup_representation.get("id")
+        if id_name and not id_name.startswith("user-content-"):
+            element_soup_representation["id"] = "user-content-" + id_name
     for link_soup_representation in html_soup.find_all("a"):
+        # "user-content-"-ify the href-attributes
         link_location = link_soup_representation.get("href")
-        if link_location.startswith("#"):
+        if link_location and link_location.startswith("#") and not link_location.startswith("#user-content-"):
+            # ^ GitHub technically doesn't recognize the last point, but we derive from GitHub's behavior here.
             link_location = "#user-content-" + link_location[1:]
             if not (link_soup_representation.has_attr("class") and link_soup_representation["class"] == ["anchor"]):
                 contains_file_internal_links = True
         link_soup_representation["href"] = link_location
+        # "user-content-"-ify the id-attribute; note that this has precedence over the name attributes.
+        link_id = link_soup_representation.get("id")
+        if link_id:
+            link_soup_representation["name"] = link_id
+            del link_id
+        # "user-content-"-ify the name-attributes
+        link_name = link_soup_representation.get("name")
+        if link_name and not link_name.startswith("user-content-"):
+            link_soup_representation["name"] = "user-content-" + link_name
     html_rendered = html_soup.__str__()
 
     # add correct id to all headings:
@@ -952,6 +1171,15 @@ def main(md_origin, origin_type="file", website_root=None, destination=None, ima
         with open(os.path.join(abs_css_paths, "code-navigation-banner-illo.svg"), "w") as to_f:
             to_f.write(from_f.read())
 
+    # check whether html is valid if we have the necessary dependency installed.
+    if imported_tidylib and validate_html:
+        _, errors = tidylib.tidy_document(html_rendered, options={'numeric-entities': 1})
+        if errors:
+            warnings.warn("The generated HTML is not entirely valid. This should not be an issue, but you ca still\n"
+                          + "raise an issue on GitHub for it"
+                          + "(https://github.com/phseiff/github-flavored-markdown-to-html/issues).\n"
+                          + str(errors))
+
     # save html where we want it to be:
     file_name_origin = md_origin.split("/")[-1].split(os.sep)[-1].rsplit(".", 1)[0]
     if "<name>" in output_name and origin_type == "string":
@@ -1218,6 +1446,16 @@ def cmd_to_main():
     done.
     """)
 
+    parser.add_argument("--emoji-support", type=int, default=1, help="""
+    Describes which level of emoji shortcode support to use. The available levels are:
+    * 1: The default. Allows the use of emoji shortcodes, e.g. `:thumbs_up:` as a shorthand for `👍️`, comparable to what
+      Discord, Telegram & Co. are doing.
+    * 0: Disable emoji shortcodes.
+    * 2: Enables emoji shortcodes, and additionally allows the use of custom emojis, by adding a link to an image in
+      between two colons (e.g. `:image.png:` will add image.png downscaled to emoji size into the text). These custom
+      emojis are properly affected by the --compress-images option, scaled to a pixel height of max. 128px, and
+      displayed with the same height as the surrounding text.""")
+
     parser.add_argument('-b', '--box-width', nargs="+", action=FuseInputString, help="""
     The text of the rendered file is always displayed in a box, like GitHub READMEs and issues are.
     By default, this box fills the entire screen (max-width: 100%%),
@@ -1271,7 +1509,7 @@ def cmd_to_main():
     help_text = "\n".join(help_text_lines)
 
     try:
-        with open_local("help.txt", "w") as help_file:
+        with open_local("help.txt", "w", encoding='utf-8') as help_file:
             help_file.write(help_text)
     except OSError:
         pass  # running an installation installed with sudo.
